/*
 * $Id: sslserver.c,v 1.2 2001/10/11 05:10:38 carzanig Exp carzanig $
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <string.h>

#include <openssl/ssl.h>

#define HOME		"./"
#define CERTF		HOME "server_cert.pem"
#define KEYF		HOME "server_priv.pem"

#define S_PORT		1111

#define BUFSIZE		4096

#define CHK_NULL(x) if ((x)==NULL) exit (1)
#define CHK_ERR(err,s) if ((err)<0) { perror(s); exit(1); }
#define CHK_SSL_ERR(err) if((err)==-1){ERR_print_errors_fp(stderr);exit(2);}

const char * Reply = "ciao Anto!\n";

int open_port(const unsigned short port)
{  
    int fd;
    struct sockaddr_in addr;
    int on = 1; /* for address reuse */
  
    fd = socket(PF_INET, SOCK_STREAM, 0);
    if (fd < 0) return -1;

    setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on));

    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = INADDR_ANY;
    addr.sin_port = htons(port);
  
    if (bind(fd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
	close(fd);
	return -1;
    }
    if (listen(fd, SOMAXCONN) < 0) {
	close(fd);
	return -1;
    }
    return fd;
}

int set_cerificate_and_key(SSL_CTX * context, 
			   const char * cfile, const char * kfile) {
    if (SSL_CTX_use_certificate_file(context, cfile, SSL_FILETYPE_PEM) <= 0)
	return -1;

    if (SSL_CTX_use_PrivateKey_file(context, kfile, SSL_FILETYPE_PEM) <= 0) 
	return -1;

    if (!SSL_CTX_check_private_key(context)) {
	fprintf(stderr, 
		"Private key does not match the certificate public key\n");
	return -2;
    }
    return 1;
}

int main()
{
    SSL_CTX *		context		= NULL;
    SSL_METHOD *	method		= NULL;
    SSL *		connection	= NULL;

    int			port_fd		= -1;
    int			sock_fd		= -1;
    int			res		= -1;
    char		buf[BUFSIZE];
    /* 
     * SSL initialization
     */
    SSL_load_error_strings();
    OpenSSL_add_ssl_algorithms();
    method = SSLv3_server_method();
    context = SSL_CTX_new(method);			CHK_NULL(context);
    /*
     * certificate and key setup
     */
    res = set_cerificate_and_key(context, CERTF, KEYF);	CHK_SSL_ERR(res);
    /*
     * TCP server setup
     */
    port_fd = open_port(S_PORT);		CHK_ERR(port_fd, "port");
    printf("+ created TCP port\n");
    sock_fd = accept(port_fd, NULL, NULL);	CHK_ERR(sock_fd, "accept");
    close(port_fd);
    printf("+ accepted TCP connection\n");
    /*
     * TCP connection is ready. Do server side SSL.
     */
    connection = SSL_new(context);		CHK_NULL(connection);
    SSL_set_fd(connection, sock_fd);
    res = SSL_accept(connection);		CHK_SSL_ERR(res);
    printf("+ performed SSL Handshake Protocol\n");
    /* 
     * Data
     */
    res = SSL_read(connection, buf, BUFSIZE - 1);	CHK_SSL_ERR(res);
    buf[res] = '\0';
    printf("+ client request:\n%s", buf);
    res = SSL_write(connection, Reply, strlen(Reply));	CHK_SSL_ERR(res);
    printf("+ my reply:\n%s", Reply);
    /* 
     * shutdown and clean up. 
     */
    close(sock_fd);
    SSL_free(connection);
    SSL_CTX_free(context);

    return 0;
}

