/*
 * $Id: sslclient.c,v 1.3 2001/10/11 05:10:43 carzanig Exp carzanig $
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#include <openssl/ssl.h>

#define S_PORT	1111
#define S_HOST	"localhost"

#define BUFSIZE	4096

#define CHK_NULL(x) if ((x)==NULL) exit (1)
#define CHK_ERR(err,s) if ((err)<0) { perror(s); exit(1); }
#define CHK_SSL_ERR(err) if((err)<0){ERR_print_errors_fp(stderr);exit(2);}

#define CAHOME		"./demoCA/"
#define CAFILE		CAHOME "cacert.pem"

#define HOME		"./"
#define CERTF		HOME "client_cert.pem"
#define KEYF		HOME "client_priv.pem"
/*
 * client request
 */
const char * Req = "ciao mamma!\n";

int connect_host_port(const char* host, unsigned short port)
{
    int fd;
    struct sockaddr_in addr;
    struct sockaddr *mapped_addr;

    fd = socket(PF_INET, SOCK_STREAM, 0);
    if (fd < 0) return -1;
  
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    if (isdigit((int)*host)) {			/* if the first char of host */
	addr.sin_addr.s_addr = inet_addr(host); /* is a digit, then assumes  */
    } else {					/*  it is the IP address     */
	struct hostent *hostEntry;
	if ((hostEntry = gethostbyname(host)) != 0) {
	    memcpy(&addr.sin_addr, hostEntry->h_addr, hostEntry->h_length);
	} else {
	    close(fd);
	    return -1;
	}
    }
    if (connect(fd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
	close(fd);
	return -1;
    }
    return fd;
}

int verify_connection(SSL * connection) 
{
    X509 *	server_cert	= NULL;
    char *	str		= NULL;
    /* 
     * Get the cipher
     */
    printf("+ SSL connection is using cipher: %s\n", 
	   SSL_get_cipher(connection));
    /*  
     * Get client's certificate 
     */
    server_cert = SSL_get_peer_certificate(connection);
    if (server_cert == NULL) {
	printf("+ server doesn't have a certificate!\n");
    } else {
	printf ("+ server certificate:\n");
    
	str = X509_NAME_oneline(X509_get_subject_name(server_cert), 0, 0);
							CHK_NULL(str);
	printf("  subject: %s\n", str);
							free(str);
    
	str = X509_NAME_oneline(X509_get_issuer_name(server_cert), 0, 0);
							CHK_NULL(str);
	printf("  issuer: %s\n", str);
							free(str);
	/* 
	 * Here's where we could do all sorts of certificate
	 * verification (before deallocating the certificate.)
	 */
    	X509_free(server_cert);
    }
}

int verify_certificate_with_ca(int ok, X509_STORE_CTX *ctx) {
    char buf[256];
    X509 *err_cert;
    int err;

    printf("+ verifying certificate within SSL Handshake\n");
    err_cert = X509_STORE_CTX_get_current_cert(ctx);
    err = X509_STORE_CTX_get_error(ctx);
    X509_NAME_oneline(X509_get_subject_name(err_cert), buf, 256);
    printf("  subject = %s\n", buf);
    if (ok) {
	printf("  verification succeeded!\n");
    } else {
	printf("  verify error: %s\n", X509_verify_cert_error_string(err));
    }
    switch (ctx->error) {
    case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
	X509_NAME_oneline(X509_get_issuer_name(ctx->current_cert),buf,256);
	printf("  issuer = %s\n",buf);
	break;
    case X509_V_ERR_CERT_NOT_YET_VALID:
    case X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD:
	printf("  not yet valid!\n");
	break;
    case X509_V_ERR_CERT_HAS_EXPIRED:
    case X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD:
	printf("  expired!\n");
	break;
    }
    return 1;
}

int set_cerificate_and_key(SSL_CTX * context, 
			   const char * cfile, const char * kfile) {
    if (SSL_CTX_use_certificate_file(context, cfile, SSL_FILETYPE_PEM) <= 0)
	return -1;

    if (SSL_CTX_use_PrivateKey_file(context, kfile, SSL_FILETYPE_PEM) <= 0) 
	return -1;

    if (!SSL_CTX_check_private_key(context)) {
	fprintf(stderr, 
		"Private key does not match the certificate public key\n");
	return -2;
    }
    return 1;
}

int main(int argc, char *argv[]) 
{
    SSL_CTX *		context		= NULL;
    SSL_METHOD *	method		= NULL;
    SSL *		connection	= NULL;

    int			sock_fd		= -1;
    int			res		= -1;
    char		buf[BUFSIZE];
    /*
     * SSL initialization
     */
    SSL_library_init();
    SSL_load_error_strings();
    OpenSSL_add_ssl_algorithms();
    method = SSLv3_client_method();
    context = SSL_CTX_new(method);
    /*
     * certificate and key setup
     */
    res = set_cerificate_and_key(context, CERTF, KEYF);	CHK_SSL_ERR(res);

    SSL_CTX_set_verify(context, SSL_VERIFY_PEER, verify_certificate_with_ca);

    SSL_CTX_set_client_CA_list(context, SSL_load_client_CA_file(CAFILE));

    if ((!SSL_CTX_load_verify_locations(context, CAFILE, NULL)) ||
	(!SSL_CTX_set_default_verify_paths(context))) {
	ERR_print_errors_fp(stdout);
    }
    /*
     * TCP connection setup
     */
    sock_fd = connect_host_port(S_HOST, S_PORT);	CHK_ERR(sock_fd, 
								"connect");
    printf("+ established TCP connection\n");
    /*
     * SSL connection setup
     */
    connection = SSL_new(context);			CHK_NULL(connection);
    res = SSL_set_fd(connection, sock_fd);		CHK_SSL_ERR(res);
    res = SSL_connect(connection);			CHK_SSL_ERR(res);
    printf("+ performed SSL Handshake Protocol\n");
    /*
     * simple connectionverification
     */
    res = verify_connection(connection);		CHK_SSL_ERR(res);

    /*
     * Data 
     */
    res = SSL_write(connection, Req, strlen(Req));	CHK_SSL_ERR(res);
    printf("+ my request:\n%s", Req);
    res = SSL_read(connection, buf, BUFSIZE - 1);	CHK_SSL_ERR(res);
    buf[res] = '\0';
    printf("+ server reply:\n%s", buf);
    /*
     * shutdown and cleanup
     */
    SSL_shutdown(connection);
    close(sock_fd);
    SSL_free(connection);
    SSL_CTX_free(context);

    return 0;
}

